import subprocess
from gplas.scripts.utils import quit_tool

def run_plasmidCC(infile, sample, minlen, species, custom_db_path,outdir):
        if species:
            cmd = f"plasmidCC -i {infile} -o {outdir}/plasmidCC -n {sample} -s {species} -l {minlen} -D -g -f"
        elif custom_db_path:
            cmd = f"plasmidCC -i {infile} -o {outdir}/plasmidCC -n {sample} -p {custom_db_path} -l {minlen} -D -g -f"

        try:
            subprocess.run(cmd, shell=True, check=True)
        except subprocess.CalledProcessError as err:
            print(f"plasmidCC returned non-zero exit status: {err.returncode}")
            quit_tool(err.returncode)


def print_speciesopts():
        cmd = "plasmidCC --speciesopts"
        try:
            subprocess.run(cmd, shell=True, check=True)
        except subprocess.CalledProcessError as err:
            print(f"plasmidCC returned non-zero exit status: {err.returncode}")
            quit_tool(err.returncode)



import pandas as pd
import igraph as ig
from Bio.SeqIO.FastaIO import SimpleFastaParser
import warnings


def scalar1(x):
    denominator = (sum([value*value for value in x]))**0.5
    scaled_x = [value/denominator for value in x]
    return scaled_x


#Create function to run three community detection algorithms on the plasmidome network.
#This fuction creates a dataframe as an output that contains the different algorithms names and the global modularity values of the resulting network.
def partitioning_components(graph):
    # Walktrap algorithm
    dendrogram_walktrap = graph.community_walktrap()
    clusters_walktrap = dendrogram_walktrap.as_clustering()
    modularity_walktrap = clusters_walktrap.modularity

    # Leading eigen values
    # TODO find a better way to fix non-converging eigenvalue error
    # For some samples there is a ~10% chance for the eigenvalue to not converge and raise an error message
    # We use this code to catch this error and/or a runtime warning and simply try the algorithm again up to 5 times
    # It is not a pretty fix but it seems to work for now
    warnings.filterwarnings("error") # allow warnings to be caught as exceptions
    for attempt in range(5):
        try:
            clusters_eigen = graph.community_leading_eigenvector()
            break
        except (ig._igraph.InternalError, RuntimeWarning):
            pass
    warnings.resetwarnings() # return warnings to normal behavior
    modularity_eigen = clusters_eigen.modularity

    # Louvain method
    clusters_louvain = graph.community_multilevel()
    modularity_louvain = clusters_louvain.modularity

    partition_info = pd.DataFrame(data={'Algorithm':['Walktrap', 'Leading-eigen', 'Louvain'],
                                        'Modularity':[modularity_walktrap, modularity_eigen, modularity_louvain]})
    return partition_info


#TODO we have multiple loops with "for row in range(solutions.shape[0]):" can we possibly merge some of them?
def calculate_coocurrence(sample, number_iterations, pred_threshold, modularity_threshold, outdir, mode='normal'):
    if mode == 'normal':
        subdir = 'normal_mode/'
    elif mode == 'unbinned':
        subdir = ''
    #Inputs
    path_nodes = f"{outdir}/gplas_input/{sample}_raw_nodes.fasta"
    path_prediction = f"{outdir}/coverage/{sample}_clean_prediction.tab"
    path_graph_repeats = f"{outdir}/coverage/{sample}_repeats_graph.tab"
    path_isolated_nodes = f"{outdir}/coverage/{sample}_isolated_nodes.tab"
    input_solutions = f"{outdir}/walks/{subdir}{sample}_solutions.tab"

    #Outputs
    output_dir = f"{outdir}/results/{subdir}"
    output_results = f"{outdir}/results/{subdir}{sample}_results_no_repeats.tab"
    output_components = f"{outdir}/results/{subdir}{sample}_bins_no_repeats.tab"
    output_png = f"{outdir}/results/{subdir}{sample}_plasmidome_network.png"

    clean_pred = pd.read_csv(path_prediction, sep='\t', header=0)
    clean_pred = clean_pred.astype({'Prob_Chromosome':float,
                                    'Prob_Plasmid':float,
                                    'Prediction':str,
                                    'Contig_name':str,
                                    'Contig_length':int,
                                    'number':str,
                                    'length':int,
                                    'coverage':float})

    repeats = pd.read_csv(path_graph_repeats, sep='\t', header=0)
    repeats.loc[:,'number'] = [name.replace('+','') for name in repeats['number']]
    repeats.loc[:,'number'] = [name.replace('-','') for name in repeats['number']]

    with open(input_solutions) as file:
        max_nodes = max([line.count('\t')+1 for line in file])

    solutions = pd.read_csv(input_solutions, sep='\t', header=None, names=range(max_nodes), engine='python')
    #TODO if the line above does not define engine='python', it produces a pandas ParserError. but only for 2 of the 214 ecoli test samples?????
    #this is not a problem anywhere else in the gplas scripts or for any other test samples
    #GCA_013602835.1_ASM1360283v1
    #GCA_013802065.1_ASM1380206v1
    #ParserError: Error tokenizing data. C error: Buffer overflow caught - possible malformed input file.
    
    #create a list with all the nodes that appear in the plasmid walks.
    all_nodes = []
    for row in range(solutions.shape[0]):
        nodes = [node for node in solutions.loc[row,:].dropna()]
        all_nodes.extend(nodes)

    #get unique set of nodes
    unique_nodes = sorted(list(set(all_nodes)))

    #TODO turn co-ocurrence matrix creation into a seperate function? moving code to functions might be useful for reusing/importing code in coocurrence_repeats
    #CREATE A CO-OCURRENCE MATRIX
    ##Column names are the nodes included in plasmid walks.
    ##Each row is a new walk
    ##Assign True if the node is present in walk and False if node is not present
    co_ocurrence = []
    for row in range(solutions.shape[0]):
        walk = [node for node in solutions.loc[row,:].dropna()]
        presence_absence = [node in walk for node in unique_nodes]
        co_ocurrence.append(presence_absence)
    co_ocurrence = pd.DataFrame(co_ocurrence, columns=unique_nodes)

    starting_nodes = [node for node in unique_nodes if node in solutions.loc[:,0].values]

    #create a dataframe for co-ocurrence frequency (in network format)
    #Start_node, Connecting_node, nr_occurences
    total_pairs = []
    #Get the number of times that two nodes co-ocuur in every walk
    for node in starting_nodes:
        index_col = [col_name == node for col_name in unique_nodes] # select the column of the target node
        index_walks = co_ocurrence.loc[:,index_col].values # select all walks where the target node is present
        walks = co_ocurrence.loc[index_walks,:].copy()
        col_sums = [sum(walks.loc[:,col]) for col in walks] # count how often any node is present in all selected walks
        for index_connecting_node in range(len(col_sums)): # save coocurrence data for every node/connecting_node
            connecting_node = unique_nodes[index_connecting_node]
            if connecting_node != node:
                weight = col_sums[index_connecting_node]
                total_pairs.append([node, connecting_node, weight])

    #Find circular sequences
    circular_sequences = []
    #Extract the walks in which start-node and end-node are the same.
    for row in range(solutions.shape[0]):
        walk = [node for node in solutions.loc[row,:].dropna()]
        if len(walk) > 1 and walk[0] == walk[-1]:
            circular_sequences.append([walk[0], walk[-1]])

    #check if the number of circular walks starting from each node equals the number of iterations.
    #if this is the case, add the circular walk to total_pairs
    if len(circular_sequences) > 0:
        no_duplicated = [sorted(list(unique_walk)) for unique_walk in set(tuple(walk) for walk in circular_sequences)] # remove duplicate entries

        for combination in range(len(no_duplicated)):
            combi = no_duplicated[combination]
            total_ocurrences = sum([walk[1] == combi[1] for walk in circular_sequences])
            if total_ocurrences == number_iterations:
                total_pairs.append([combi[0], combi[1], total_ocurrences])

    total_pairs = pd.DataFrame(total_pairs, columns=['Starting_node', 'Connecting_node', 'weight'])

    total_pairs.loc[:,'Starting_node'] = [node.replace('+','') for node in total_pairs.loc[:,'Starting_node']]
    total_pairs.loc[:,'Starting_node'] = [node.replace('-','') for node in total_pairs.loc[:,'Starting_node']]
    total_pairs.loc[:,'Connecting_node'] = [node.replace('+','') for node in total_pairs.loc[:,'Connecting_node']]
    total_pairs.loc[:,'Connecting_node'] = [node.replace('-','') for node in total_pairs.loc[:,'Connecting_node']]

    #Filter-out cases of no-coocurrence
    #TODO ASK Julian: we filter on weight > 1 why not weight > 0?
    index = [weight > 1 for weight in total_pairs.loc[:,'weight']]
    total_pairs = total_pairs.loc[index,:]

    #Scale weights
    complete_node_info = pd.DataFrame(columns=['Starting_node', 'Connecting_node', 'weight'])
    for node in sorted(list(set(total_pairs.loc[:,'Starting_node']))):
        index = total_pairs.loc[:,'Starting_node'] == node
        first_node = total_pairs.loc[index,:]
        particular_node = []

        for connecting_node in sorted(list(set(first_node.loc[:,'Connecting_node']))):
            index = first_node.loc[:,'Connecting_node'] == connecting_node
            first_second_nodes = first_node.loc[index,:]
            total_weight = sum(first_second_nodes.loc[:,'weight'])
            particular_node.append([node, connecting_node, total_weight])

        particular_node = pd.DataFrame(particular_node, columns=['Starting_node', 'Connecting_node', 'weight'])
        particular_node.loc[:,'scaled_weight'] = scalar1(particular_node.loc[:,'weight']) # add a column with scaled weights using scalar1()
        complete_node_info = pd.concat([complete_node_info, particular_node], ignore_index=True)

    total_pairs = complete_node_info

    initial_nodes = [node.replace('+','') for node in starting_nodes]
    initial_nodes = [node.replace('-','') for node in initial_nodes]

    #Filter out repeated elements. Keep only connections of unitigs.
    index = ((total_pairs.loc[:,'Starting_node'].isin(initial_nodes)) &
             (total_pairs.loc[:,'Connecting_node'].isin(initial_nodes)))
    total_pairs = total_pairs.loc[index,:]

    #Reformat the dataframe to obtain the totality of co-courences
    weight_counting = []
    #First check if we actually have co-ocurrence of unitigs.
    if total_pairs.shape[0] > 0 and total_pairs.shape[1] > 0:
        for row in range(total_pairs.shape[0]):
            initial_node = total_pairs.iloc[row,0]
            connecting_node = total_pairs.iloc[row,1]
            raw_count = int(total_pairs.iloc[row,2])
            if int(initial_node) < int(connecting_node):
                pair = f"{initial_node}-{connecting_node}"
            else:
                pair = f"{connecting_node}-{initial_node}"

            weight_counting.append([pair, raw_count])
    else:
        index = clean_pred.loc[:,'Prob_Plasmid'] > pred_threshold
        pl_unbinned = clean_pred.loc[index,:].copy()
        pl_unbinned = pl_unbinned.drop(columns='Contig_length')
        pl_unbinned['Component'] = pd.Series(dtype='object')
        pl_unbinned['Prob_Chromosome'] = pd.Series(dtype='float')
        pl_unbinned['Prob_Plasmid'] = pd.Series(dtype='float')
        pl_unbinned['coverage'] = pd.Series(dtype='float')
        if pl_unbinned.shape[0] > 0:
            pl_unbinned.loc[:,'Component'] = 'Unbinned'
            pl_unbinned.loc[:,'Prob_Chromosome'] = round(pl_unbinned.loc[:,'Prob_Chromosome'], 2)
            pl_unbinned.loc[:,'Prob_Plasmid'] = round(pl_unbinned.loc[:,'Prob_Plasmid'], 2)
            pl_unbinned.loc[:,'coverage'] = round(pl_unbinned.loc[:,'coverage'], 2)

        with open(path_nodes) as file:
            raw_nodes = [[str(values[0]), str(values[1])] for values in SimpleFastaParser(file)]

        df_nodes = pd.DataFrame(data=raw_nodes, columns=['Contig_name', 'Sequence'])
        df_nodes = df_nodes.merge(pl_unbinned, on='Contig_name')

        for component in sorted(list(set(df_nodes.loc[:,'Component']))):
            index = df_nodes.loc[:,'Component'] == component
            nodes_component = df_nodes.loc[index,:]
            component_complete_name = f"{sample}_bin_{component}"
            filename = f"{output_dir}{component_complete_name}.fasta"

            with open(filename, mode='w') as file:
                for contig in range(nodes_component.shape[0]):
                    file.write('>' + nodes_component.iloc[contig,0] + '\n' + nodes_component.iloc[contig,1] + '\n')
        #TODO if possible, name the column 'Bin' from the start instead of 'Component' and then renaming it later
        pl_unbinned = pl_unbinned.rename(columns={'Component':'Bin'})
        results_subgraph = pl_unbinned.loc[:,['number','Bin']]

        pl_unbinned.to_csv(output_results, sep='\t', index=False, header=True, mode='w')
        results_subgraph.to_csv(output_components, sep='\t', index=False, header=True, mode='w')
        return False

    weight_counting = pd.DataFrame(weight_counting, columns=['Pair','Count'])

    unique_pairs = sorted(list(set(weight_counting.loc[:,'Pair'])))
    sum_weights = []
    for pair in unique_pairs:
        index = list(weight_counting.loc[:,'Pair'] == pair)
        sum_weights.append(sum(weight_counting.loc[index,'Count']))

    pairs = [pair.split('-') for pair in unique_pairs]

    weight_graph = pd.DataFrame(data={'From_to':[pair[0] for pair in pairs],
                                      'To_from':[pair[1] for pair in pairs],
                                      'weight':sum_weights})

    full_graph_info = pd.DataFrame()

    for node in sorted(list(set(weight_graph.loc[:,'From_to']))):
        index = weight_graph.loc[:,'From_to'] == node
        df_node = weight_graph.loc[index,:].copy()
        df_node.loc[:,'scaled_weight'] = scalar1(df_node.loc[:,'weight'])
        full_graph_info = pd.concat([full_graph_info, df_node], ignore_index=True)

    full_graph_info.loc[:,'width'] = full_graph_info.loc[:,'scaled_weight'].copy() * 5

    graph_pairs = ig.Graph.DataFrame(full_graph_info, directed=False, use_vids=False)

    # Simplifying the graph - Removing self-loops from the graph
    no_loops_graph = graph_pairs.simplify(multiple=False)

    #Analyze if the plasmidome network can be partitioned into different sub-networks.
    #Get clusters based on connectivity alone.
    components_graph = no_loops_graph.decompose(mode='weak', minelements=2)
    
    #Get a table that contains each node and the component it belongs to
    components = no_loops_graph.connected_components()
    info_comp_member = components.membership
    original_components = sorted(list(set(info_comp_member)))
    info_comp_size = components.sizes()

    node_and_component = pd.DataFrame(data={'Node':[no_loops_graph.vs[index]['name'] for index in range(len(no_loops_graph.vs))],
                                            'Original_component':info_comp_member})

    information_components = pd.DataFrame(data={'Original_component':original_components,
                                                'Size':info_comp_size})

    full_info_components = node_and_component.merge(information_components, on='Original_component')

    #Analyze if each component needs to be sub-divided
    #Create a data-frame to hold the results from the different clusters
    complete_partition_info = pd.DataFrame()
    #TODO this statement is always true?? when is len(components_graph) ever 0? can we remove this statement?
    if len(components_graph) >= 1:
        #Loop through each component and run the different community detection algorithms.
        #As output we get a table with the different modularity values of each community detection algorithm for each sub-graph.
        for component in range(len(components_graph)):
            subgraph = components_graph[component]
            partition_info = partitioning_components(subgraph)
            first_node = subgraph.vs[0]['name']
            index = full_info_components.loc[:,'Node'] == first_node
            info_first_node = full_info_components.loc[index,:]

            partition_info.loc[:,'Original_component'] = info_first_node.loc[:,'Original_component'].values[0]
            complete_partition_info = pd.concat([complete_partition_info, partition_info], ignore_index=True)

        complete_partition_info.loc[:,'Modularity'] = round(complete_partition_info.loc[:,'Modularity'], 2)

        #Get the decision of splitting or not. if the modularity value is bigger than the threshold, split. Otherwise, don't.
        complete_partition_info.loc[:,'Decision'] = ['Split' if score >= modularity_threshold else 'No_split' for score in complete_partition_info.loc[:,'Modularity']]

    #Add singletons components, namely Single-node components from the plasmidome network.
    index = [info_comp_size[component] == 1 for component in range(len(info_comp_size))]
    singletons_component = [i for i, x in enumerate(index) if x]  # TODO replace this list comprehension with np.where?

    #Add the singletons to the rest of the data.
    if len(singletons_component) > 0:
        df_singletons = pd.DataFrame(data={'Algorithm':'Independent_single_component',
                                           'Modularity':0,
                                           'Original_component':singletons_component,
                                           'Decision':'No_split'})
        complete_partition_info = pd.concat([complete_partition_info, df_singletons], ignore_index=True)

    complete_partition_info.sort_values(by='Original_component', axis=0, ascending=True, inplace=True, ignore_index=True)

    #When suitable, get the results from the network partitioning algorithm.
    contigs_membership = []

    internal_component = 0

    #For determining partition, get the algorithm that provides the biggest modularity value.
    for component in sorted(list(set(complete_partition_info.loc[:,'Original_component']))):
        index = complete_partition_info.loc[:,'Original_component'] == component
        decision_comp = complete_partition_info.loc[index,:]

        if decision_comp.iloc[0,0] != 'Independent_single_component':
            split_decision = sum(decision_comp.loc[:,'Decision'].str.count('Split'))
            no_split_decision = sum(decision_comp.loc[:,'Decision'].str.count('No_split'))

            if split_decision >= no_split_decision:
                index = decision_comp.loc[:,'Modularity'] == max(decision_comp.loc[:,'Modularity'])
                algorithm_to_split = decision_comp.loc[index,:]
                algorithm = algorithm_to_split.iloc[0,0]

                graph_component = components_graph[internal_component]
                spl_names = graph_component.vs['name']

                if algorithm == 'Walktrap':
                    dendrogram_walktrap = graph_component.community_walktrap()
                    clusters_walktrap = dendrogram_walktrap.as_clustering()
                    spl_membership = clusters_walktrap.membership

                elif algorithm == 'Leading-eigen':
                    clusters_eigen = graph_component.community_leading_eigenvector()
                    spl_membership = clusters_eigen.membership

                elif algorithm == 'Louvain':
                    clusters_louvain = graph_component.community_multilevel()
                    spl_membership = clusters_louvain.membership

                for node in range(len(spl_names)):
                    contigs_membership.append([algorithm, component, spl_membership[node], spl_names[node]])

            else:
                index = full_info_components.loc[:,'Original_component'] == component
                nodes_component = full_info_components.loc[index,:]

                for node in range(nodes_component.shape[0]): #TODO check if this 0 from membership is always 0?
                    contigs_membership.append(['Not_split_component', component, 0, nodes_component.iloc[node,0]])

            internal_component += 1

        else:
            index = full_info_components.loc[:,'Original_component'] == component
            nodes_component = full_info_components.loc[index,:]

            for node in range(nodes_component.shape[0]):
                contigs_membership.append(['Independent_single_component', component, 0, nodes_component.iloc[node,0]])

    contigs_membership = pd.DataFrame(data=contigs_membership, columns=['Algorithm','Original_component','Bin','Contig'])
    contigs_membership.loc[:,'Cluster'] = ['-'.join([str(contigs_membership.loc[row,'Original_component']), str(contigs_membership.loc[row,'Bin'])]) for row in range(contigs_membership.shape[0])]
    contigs_membership.sort_values(by='Cluster', axis=0, ascending=True, inplace=True, ignore_index=True)
    contigs_membership.loc[:,'Final_cluster'] = contigs_membership.loc[:,'Cluster'].copy()

    unique_clusters = sorted(list(set(contigs_membership.loc[:,'Cluster'])))
    for number in range(len(unique_clusters)):
        cluster_name = unique_clusters[number]
        contigs_membership.loc[:,'Final_cluster'] = [cluster.replace(cluster_name, str(number)) for cluster in contigs_membership.loc[:,'Final_cluster']]

    #TODO what happens if there are more than 31 clusters and it runs out of hardcoded colors?
    set_colors = ['#add8e6','#d49f36','#507f2d','#84b67c','#a06fda','#df462a','#5a51dc','#5b83db','#c76c2d','#4f49a3','#552095','#82702d','#dd6bbb','#334c22','#d83979','#55baad','#dc4555','#62aad3','#8c3025','#417d61','#862977','#bba672','#403367','#da8a6d','#a79cd4','#71482c','#c689d0','#6b2940','#d593a7','#895c8b','#bd5975']
    contigs_membership.loc[:,'Color'] = [set_colors[int(cluster)] for cluster in contigs_membership.loc[:,'Final_cluster']]
    order_contigs = pd.DataFrame(data=no_loops_graph.vs['name'], columns=['Contig'])
    contigs_membership = order_contigs.merge(contigs_membership, on='Contig')
    no_loops_graph.vs['color'] = contigs_membership.loc[:,'Color']

    ig.plot(no_loops_graph,
            target=output_png,
            bbox=(700,700),
            margin=50,
            vertex_size=40,
            vertex_label=no_loops_graph.vs['name'],
            edge_width=no_loops_graph.es['width'],
            edge_color='grey')

    #Get the final membership data after partitioning
    results_subgraph = pd.DataFrame(data={'number':contigs_membership.loc[:,'Contig'],
                                          'Component':contigs_membership.loc[:,'Final_cluster']})

    #Get all the plasmid nodes
    index = clean_pred.loc[:,'Prob_Plasmid'] >= pred_threshold
    pl_nodes = clean_pred.loc[index,:].copy() # Selecting only contigs predicted as plasmid-derived

    #Get all the not-assigned nodes (Unbinned and repeats)
    index = ~pl_nodes.loc[:,'number'].isin(results_subgraph.loc[:,'number'])
    pl_notassigned = pl_nodes.loc[index,:].copy()
    #Get the repeats
    index = pl_notassigned.loc[:,'number'].isin(repeats.loc[:,'number'])
    pl_repeats = pl_notassigned.loc[index,:].copy()
    #Get the Unbinned nodes
    index = ~pl_notassigned.loc[:,'number'].isin(repeats.loc[:,'number'])
    pl_unbinned = pl_notassigned.loc[index,:].copy()

    #If there are isolated nodes, remove them from the unbinned and create a new category
    isolated_nodes = pd.read_csv(path_isolated_nodes, sep='\t', header=0)
    isolated_nodes = isolated_nodes.astype({'Prob_Chromosome':float,
                                            'Prob_Plasmid':float,
                                            'Prediction':str,
                                            'Contig_name':str,
                                            'Contig_length':int,
                                            'number':str,
                                            'length':int,
                                            'coverage':float})
    
    index = pl_unbinned.loc[:,'number'].isin(isolated_nodes.loc[:,'number'])
    pl_isolated = pl_unbinned.loc[index,:].copy()
    index = ~pl_unbinned.loc[:,'number'].isin(isolated_nodes.loc[:,'number'])
    pl_unbinned = pl_unbinned.loc[index,:]

    #Get the assigned nodes based on the final membership algorithm
    index = pl_nodes.loc[:,'number'].isin(results_subgraph.loc[:,'number'])
    pl_assigned = pl_nodes.loc[index,:]

    #Get all the information from the plasmid nodes (Lenght, coverage, bin number, etc)
    full_info_assigned = pl_assigned.merge(results_subgraph, on='number')

    #Add unbinned category
    if pl_unbinned.shape[0] >= 1:
        pl_unbinned.loc[:,'Component'] = 'Unbinned'
        full_info_assigned = pd.concat([full_info_assigned, pl_unbinned], ignore_index=True)

    #Add isolated nodes category
    if pl_isolated.shape[0] >= 1:
        isolated_components = []
        for isolated_nr in range(pl_isolated.shape[0]):
            isolated_components.append(f"Isolated_{isolated_nr}")
        
        pl_isolated['Component'] = isolated_components

        full_info_assigned = pd.concat([full_info_assigned, pl_isolated], ignore_index=True)

    #Add repeat-like category
    if pl_repeats.shape[0] >= 1:
        pl_repeats.loc[:,'Component'] = 'Repeat_like'
        full_info_assigned = pd.concat([full_info_assigned, pl_repeats], ignore_index=True)

    #Add information to the results file
    full_info_assigned = full_info_assigned.drop(columns='Contig_length')
    full_info_assigned.loc[:,'Prob_Chromosome'] = round(full_info_assigned.loc[:,'Prob_Chromosome'], 2)
    full_info_assigned.loc[:,'Prob_Plasmid'] = round(full_info_assigned.loc[:,'Prob_Plasmid'], 2)
    full_info_assigned.loc[:,'coverage'] = round(full_info_assigned.loc[:,'coverage'], 2)

    #Create the fasta files
    with open(path_nodes) as file:
        raw_nodes = [[str(values[0]), str(values[1])] for values in SimpleFastaParser(file)]

    df_nodes = pd.DataFrame(data=raw_nodes, columns=['Contig_name', 'Sequence'])
    df_nodes = df_nodes.merge(full_info_assigned, on='Contig_name')

    #Write fasta files
    for component in sorted(list(set(df_nodes.loc[:,'Component']))):
        index = df_nodes.loc[:,'Component'] == component
        nodes_component = df_nodes.loc[index,:]
        component_complete_name = f"{sample}_bin_{component}"
        filename = f"{output_dir}{component_complete_name}.fasta"

        with open(filename, mode='w') as file:
            for contig in range(nodes_component.shape[0]):
                file.write('>' + nodes_component.iloc[contig,0] + '\n' + nodes_component.iloc[contig,1] + '\n')

    #TODO if possible start the column name as 'Bin' instead of renaming it at the end; like in coocurrence_repeats.py
    full_info_assigned = full_info_assigned.rename(columns={'Component':'Bin'})
    results_subgraph = results_subgraph.rename(columns={'Component':'Bin'})

    full_info_assigned.to_csv(output_results, sep='\t', index=False, header=True, mode='w')
    results_subgraph.to_csv(output_components, sep='\t', index=False, header=True, mode='w')

    return True



import pandas as pd
from pandas.api.types import is_integer_dtype, is_float_dtype, is_object_dtype, is_string_dtype
from Bio.SeqIO.FastaIO import SimpleFastaParser


class PredictionFileFormatError(Exception):
    pass


def check_prediction(sample, path_prediction,outdir):
    #0. Load prediction file.
    prediction_file = pd.read_csv(path_prediction, sep='\t', header=0)

    #######################################################################################################################################

    #1. Check the number of columns
    if prediction_file.shape[1] != 5:
        raise PredictionFileFormatError("Prediction file should contain 5 tab-separated columns")

    #######################################################################################################################################

    #2. Check the column names
    if prediction_file.columns[0] != 'Prob_Chromosome':
        raise PredictionFileFormatError("First column should be named Prob_Chromosome (case sensitive)")

    if prediction_file.columns[1] != 'Prob_Plasmid':
        raise PredictionFileFormatError("Second column should be named Prob_Plasmid (case sensitive)")

    if prediction_file.columns[2] != 'Prediction':
        raise PredictionFileFormatError("Third column should be named Prediction (case sensitive)")

    if prediction_file.columns[3] != 'Contig_name':
        raise PredictionFileFormatError("Fourth column should be named Contig_name (case sensitive)")

    if prediction_file.columns[4] != 'Contig_length':
        raise PredictionFileFormatError("Fifth column should be named Contig_length (case sensitive)")

    #######################################################################################################################################

    #3. Check the data type of every column
    if(not is_float_dtype(prediction_file.dtypes['Prob_Chromosome']) and
       not is_integer_dtype(prediction_file.dtypes['Prob_Chromosome'])):
        raise PredictionFileFormatError("First column should be of type float/integer")

    if max(prediction_file['Prob_Chromosome']) > 1 or min(prediction_file['Prob_Chromosome']) < 0:
        raise PredictionFileFormatError("First column should contain values between 0 and 1")

    if(not is_float_dtype(prediction_file.dtypes['Prob_Plasmid']) and
       not is_integer_dtype(prediction_file.dtypes['Prob_Plasmid'])):
        raise PredictionFileFormatError("Second column should be of type float/integer")

    if max(prediction_file['Prob_Plasmid']) > 1 or min(prediction_file['Prob_Plasmid']) < 0:
        raise PredictionFileFormatError("Second column should contain values between 0 and 1")

    if(not is_object_dtype(prediction_file.dtypes['Prediction']) and
       not is_string_dtype(prediction_file.dtypes['Prediction'])):
        raise PredictionFileFormatError("Third column should be of type character/string")

    valid_predictions = ['Plasmid', 'Chromosome']
    if any([pred not in valid_predictions for pred in prediction_file['Prediction']]):
        raise PredictionFileFormatError(f"Third column values should be either {' or '.join(valid_predictions)} (case sensitive)")

    if(not is_object_dtype(prediction_file.dtypes['Contig_name']) and
       not is_string_dtype(prediction_file.dtypes['Contig_name'])):
        raise PredictionFileFormatError("Fourth column should be of type character/string")

    if not is_integer_dtype(prediction_file.dtypes['Contig_length']):
        raise PredictionFileFormatError("Fifth column should be of type integer")

    #######################################################################################################################################

    #4. check if plasmids exist in the prediction
    if 'Plasmid' not in prediction_file['Prediction'].values:
        raise PredictionFileFormatError("There are no plasmids in the prediction file, gplas can't do anything")

    #######################################################################################################################################

    #5. Check if the contig names in the prediction match the names in the FASTA file
    ##5.0 Get a path for fasta file.
    raw_nodes_path = f"{outdir}/gplas_input/{sample}_raw_nodes.fasta" 

    ##5.1 Get headers from fastafile.
    with open(raw_nodes_path) as file:
        fasta_headers = [str(node[0]) for node in SimpleFastaParser(file)]

    ##5.2 Get the headers from the prediction file
    prediction_headers = prediction_file['Contig_name']

    ##5.3 See if the predictions are in the fasta headers
    #TODO ASK Julian: do we also need to check the other way around? now there can be contigs in the FASTA that are not in the prediction file
    #TODO ASK Julian: should we use contigs.fasta instead of raw_nodes.fasta when checking this?
    comparison_output = [header in fasta_headers for header in prediction_headers]

    if not all(comparison_output):
        raise PredictionFileFormatError(f"Contig names in prediction file should match exactly with those in '{raw_nodes_path}'")

    #######################################################################################################################################

    #6. All checks are successful!
    return



import pandas as pd
import numpy as np

def generate_repeat_paths(sample, number_iterations, filtering_threshold,outdir):
    #Inputs
    path_links = f"{outdir}/coverage/{sample}_clean_links.tab"
    path_prediction = f"{outdir}/coverage/{sample}_clean_prediction.tab"
    path_graph_contigs = f"{outdir}/coverage/{sample}_graph_contigs.tab"
    path_graph_repeats = f"{outdir}/coverage/{sample}_repeats_graph.tab"
    path_init_nodes = f"{outdir}/coverage/{sample}_repeat_nodes.tab"
    #Params
    # TODO make these into user-tunable parameters?
    number_nodes = 20
    prob_small_repeats = 0.5
    #Outputs
    output_path = f"{outdir}/walks/repeats/{sample}_solutions.tab"

    links = pd.read_csv(path_links, sep='\t', header=None)
    clean_pred = pd.read_csv(path_prediction, sep='\t', header=0)
    clean_pred = clean_pred.astype({'Prob_Chromosome':float,
                                    'Prob_Plasmid':float,
                                    'Prediction':str,
                                    'Contig_name':str,
                                    'Contig_length':int,
                                    'number':str,
                                    'length':int,
                                    'coverage':float})

    graph_contigs = pd.read_csv(path_graph_contigs, sep='\t', header=0)

    small_contigs = graph_contigs[graph_contigs['length'] < 500].copy()
    small_contigs_signed_nodes = small_contigs.copy()
    small_contigs_signed_nodes = [node for node in small_contigs_signed_nodes['number']]
    small_contigs.loc[:,'number'] = [name.replace('+','') for name in small_contigs['number']]
    small_contigs.loc[:,'number'] = [name.replace('-','') for name in small_contigs['number']]

    repeats = pd.read_csv(path_graph_repeats, sep='\t', header=0)
    repeats_signed_nodes = repeats.copy()
    repeats_signed_nodes = [node for node in repeats_signed_nodes['number']]
    repeats.loc[:,'number'] = [name.replace('+','') for name in repeats['number']]
    repeats.loc[:,'number'] = [name.replace('-','') for name in repeats['number']]

    initialize_nodes = pd.read_csv(path_init_nodes, sep='\t', header=None)
    initialize_nodes = [str(node) for node in initialize_nodes.iloc[:,0]]

    def plasmid_graph(initial_seed, output_path, links, number_iterations, number_nodes, filtering_threshold, prob_small_repeats, classification, direction):
        paths_list = []
        for iteration in range(number_iterations): # Number of times we repeat this process
            path = [initial_seed] # We add the initial seed to the path, first element in the list
            seed = initial_seed
            unitig_seed_classification = classification

            #################################### Coverage of the current path #################################

            # Extracting the info from our current path                     
            info_path = graph_contigs[[number in path for number in graph_contigs['number']]].copy()
            length_path = sum(info_path['length']) # Length of the path
            info_path['contribution'] = info_path['length']/length_path # Shorter contigs should have a lower contribution to the coverage. k-mer coverage on these contigs fluctuates drastically
            path_mean = np.average(info_path['coverage'], weights=info_path['contribution']) # Coverage of the current path

            ##################### Elongating the path ###########################################

            for elongation in range(number_nodes):
                index = links.loc[:,0] == seed
                current_links = links.loc[index,:] # Consider the last element present in our path and observe all the possible links

                if(current_links.shape[0] == 0):
                    output = ','.join(path) #join path with ',' for later step in coocurrence_repeats script
                    output = '\t'.join([output, classification, unitig_seed_classification, str(path_mean)])
                    paths_list.append(output)
                    path = [initial_seed] # There are no connections possible from this contig
                    unitig_seed_classification = classification
                    break # Exiting elongation loop

                list_connections = sorted(list(set(current_links[2]))) # All the possible unique connections 

                #TODO ASK Julian: why is classification != 'Plasmid' also checked here?; why run this code if not plasmid but len = 1??; should the OR be changed to AND?
                # We do not allow that a node which is not a repeat appears more than 1 time in any solution but we exclude the initial seed from this consideration 
                if((len(path) > 1) | (classification != 'Plasmid')): # If the path has more than one element        
                    remove_nodes = path[1:]

                    remove_nodes = [node.replace('+','') for node in remove_nodes]
                    remove_nodes = [node.replace('-','') for node in remove_nodes]

                    positive_remove_nodes = [node + '+' for node in remove_nodes]
                    negative_remove_nodes = [node + '-' for node in remove_nodes]

                    ommit_nodes = positive_remove_nodes + negative_remove_nodes

                    first_node = path[0]

                    if(direction == 'forward'):
                        first_node_to_exclude = first_node.replace('+','-')

                    elif(direction == 'reverse'):
                        first_node_to_exclude = first_node.replace('-','+')

                    ommit_nodes.append(first_node_to_exclude)

                    # We need to remove directionality from the path to avoid paths e.g. 54-,161-,54+
                    ommit_nodes = [node for node in ommit_nodes if node not in repeats_signed_nodes]

                    list_connections = [str(connection) for connection in list_connections if connection not in ommit_nodes] # Avoiding loops within the solution

                total_connections = len(list_connections) # Number of connections
                base_probabilities = [float(0)]*total_connections # Generating a list with the connection probabilities

                if(total_connections < 1):
                    output = ','.join(path) #join path with ',' for later step in coocurrence_repeats script
                    output = '\t'.join([output, classification, unitig_seed_classification, str(path_mean)])
                    paths_list.append(output)
                    path = [initial_seed] # There are no connections possible from this contig
                    unitig_seed_classification = classification
                    break # Exiting elongation loop

                ################################# Probabilities of the connections #####################################################

                # We generate a dataframe with the connection probabilities
                prob_df = pd.DataFrame(data={'number':list_connections,
                                             'Prob_Plasmid':base_probabilities,
                                             'Prob_Chromosome':base_probabilities})

                # Replacing the base probabilities with the probabilities generated by mlplasmids 
                prob_df.loc[:,'number'] = [number.replace('+','') for number in prob_df['number']]
                prob_df.loc[:,'number'] = [number.replace('-','') for number in prob_df['number']]

                if(((classification != 'Plasmid') & (classification != 'Chromosome')) | ((unitig_seed_classification != 'Plasmid') & (unitig_seed_classification != 'Chromosome'))):
                    final_probs = pd.Series([float(0.5)]*total_connections)
                else:
                    for number in prob_df.loc[:,'number']:
                        matchID = clean_pred.index[clean_pred['number'] == number]
                        if any(matchID):
                            index = prob_df.loc[:,'number'] == number
                            prob_df.loc[index, 'Prob_Plasmid'] = float(clean_pred.loc[matchID,'Prob_Plasmid'].values[0])
                            prob_df.loc[index, 'Prob_Chromosome'] = float(clean_pred.loc[matchID,'Prob_Chromosome'].values[0])

                    if((classification == 'Plasmid') | (unitig_seed_classification == 'Plasmid')):
                        final_probs = prob_df.loc[:,'Prob_Plasmid'].copy()
                    elif((classification=='Chromosome') | (unitig_seed_classification=='Chromosome')):
                        final_probs = prob_df.loc[:,'Prob_Chromosome'].copy()

                # Short contigs do not have a reliable probability, we assign them a predefined probability (passed via the argument 'prob_small_repeats')
                index = prob_df.loc[:,'number'].isin(small_contigs.loc[:,'number'])
                final_probs.loc[index] = float(prob_small_repeats) # OVERLAP!  

                # Transposases are also corner-cases for the machine-learning algorithm, we follow the same principle as done with short contigs 
                index = prob_df.loc[:,'number'].isin(repeats.loc[:,'number'])
                final_probs.loc[index] = float(prob_small_repeats)

                record_connections = []
                for i in range(len(final_probs)):
                    record_connections.append([1.0, number_iterations, iteration, elongation, initial_seed, seed, list_connections[i], final_probs[i]])

                record_connections = pd.DataFrame(data=record_connections, columns=['factor','number_iterations','iteration','elongation','first_node','ingoing_node','outgoing_node','Probability_pl_chr'])

                index = graph_contigs.loc[:,'number'].isin(record_connections.loc[:,'outgoing_node'])
                cov_connections_info = graph_contigs.loc[index,:].copy()

                cov_connections_info.loc[:,'Probability_cov'] = float(1)
                #TODO find a better way to do this merge
                record_connections = record_connections.merge(cov_connections_info[['number','Probability_cov']], left_on='outgoing_node', right_on='number')
                record_connections = record_connections.drop(columns='number')

                index = record_connections.loc[:,'outgoing_node'].isin(repeats_signed_nodes)
                record_connections.loc[index,'Probability_cov'] = float(prob_small_repeats)

                index = record_connections.loc[:,'outgoing_node'].isin(small_contigs_signed_nodes)
                record_connections.loc[index,'Probability_cov'] = float(prob_small_repeats)                        

                record_connections.loc[:,'Probability'] = record_connections.loc[:,'Probability_pl_chr'] * record_connections.loc[:,'Probability_cov']
                record_connections.loc[:,'Probability_freq'] = record_connections.loc[:,'Probability']/sum(record_connections.loc[:,'Probability'])
                record_connections.loc[:,'Verdict'] = 'non-selected'

                if(sum(record_connections.loc[:,'Probability'] >= filtering_threshold) == 0):
                    output = ','.join(path)  # We join path with ',' for later step in coocurrence_repeats script
                    output = '\t'.join([output, classification, unitig_seed_classification, str(path_mean)])
                    paths_list.append(output)                    
                    path = [initial_seed] # There are no connections possible from this contig
                    unitig_seed_classification = classification
                    break # Exiting elongation loop

                # Filter step to avoid going into really bad connections
                index = list(record_connections.loc[:,'Probability'] >= filtering_threshold)
                filter_connections = record_connections.loc[index,:].copy()
                if(sum(filter_connections['Probability_freq']) != 1): # recalculate probability frequencies after filter
                    filter_connections['Probability_freq'] = filter_connections.loc[:,'Probability']/sum(filter_connections.loc[:,'Probability'])

                random_connection = np.random.choice(filter_connections.loc[:,'outgoing_node'], size=1, p=filter_connections.loc[:,'Probability_freq'])[0] # Choose one connection 
                path.append(str(random_connection))

                #TODO ASK Julian: why do we check for plasmid here?
                if((random_connection == path[0]) & (classification == 'Plasmid')):
                    output = ','.join(path) # We join path with ',' for a later step in coocurrence_repeats script
                    output = '\t'.join([output, classification, unitig_seed_classification, str(path_mean)])
                    paths_list.append(output)
                    path = [initial_seed]
                    unitig_seed_classification = classification
                    break # Exiting elongation loop

                seed = str(random_connection)

                #updating the info_path variable.
                index = graph_contigs.loc[:,'number'].isin(path)
                info_path = graph_contigs.loc[index,:].copy()

                #get a seed clean, with no directions. This will be used to re-classify path in case that we are starting from a Repeat
                seed_clean = seed
                seed_clean = seed_clean.replace('+','')
                seed_clean = seed_clean.replace('-','')

                #In case that we've started the walks from a repeated element, create a new classification for the path (If the last seed has one). Otherwise, keep calling it a repeat.
                if(((unitig_seed_classification == 'Repeat') | (unitig_seed_classification == 'Small')) & (sum(clean_pred.loc[:,'number'].isin([seed_clean])) > 0)):
                    unitig_seed_info = clean_pred.loc[clean_pred.loc[:,'number'].isin([seed_clean]),:]
                    unitig_seed_classification = unitig_seed_info.loc[:,'Prediction'].values[0]
                elif((unitig_seed_classification == 'Repeat') & (sum(info_path.loc[:,'number'].isin(repeats_signed_nodes)) != info_path.shape[0])):
                    unitig_seed_classification = 'Small'

                # We exit the function if we have reached the maximum number of nodes allowed per path
                if(len(path) >= number_nodes): 
                    output = ','.join(path) # We join path with ',' for later step in coocurrence_repeats script
                    output = '\t'.join([output, classification, unitig_seed_classification, str(path_mean)])
                    paths_list.append(output)
                    path = [initial_seed]
                    unitig_seed_classification = classification
                    break # Exiting elongation loop

                if(unitig_seed_classification != 'Repeat'):
                    index = ~info_path.loc[:,'number'].isin(repeats_signed_nodes)
                    info_path = info_path.loc[index,:]
                    length_path = sum(info_path.loc[:,'length'])
                    info_path.loc[:,'contribution'] = info_path.loc[:,'length'] / length_path
                    path_mean = np.average(info_path.loc[:,'coverage'], weights=info_path.loc[:,'contribution']) # Coverage of the current path

                    if((unitig_seed_classification == 'Plasmid') | (unitig_seed_classification == 'Chromosome')):
                        output = ','.join(path) # We join path with ',' for later step in coocurrence_repeats script
                        output = '\t'.join([output, classification, unitig_seed_classification, str(path_mean)])
                        paths_list.append(output)
                        path = [initial_seed]
                        unitig_seed_classification = classification
                        break # Exiting elongation loop
                else:
                    first_node = info_path.loc[info_path.loc[:,'number'].isin([initial_seed])].copy()
                    index = ~info_path.loc[:,'number'].isin(repeats_signed_nodes)
                    info_path = info_path.loc[index,:]
                    info_path = pd.concat([info_path, first_node], ignore_index=True)                
                    length_path = sum(info_path.loc[:,'length'])
                    info_path.loc[:,'contribution'] = info_path.loc[:,'length'] / length_path
                    path_mean = np.average(info_path.loc[:,'coverage'], weights=info_path.loc[:,'contribution']) # Coverage of the current path

        return paths_list


    final_paths = []
    for seed in initialize_nodes:
        np.random.seed(123)
        seed_info = clean_pred.loc[clean_pred.loc[:,'number'].isin([seed]),:]

        if(seed_info.shape[0] != 0):
            seed_classification = seed_info.loc[:,'Prediction']
        else:
            seed_classification = 'Repeat'

        positive_seed = seed + '+'
        negative_seed = seed + '-'
        final_paths.extend(plasmid_graph(positive_seed, output_path, links, number_iterations, number_nodes, filtering_threshold, prob_small_repeats, seed_classification, direction='forward'))
        final_paths.extend(plasmid_graph(negative_seed, output_path, links, number_iterations, number_nodes, filtering_threshold, prob_small_repeats, seed_classification, direction='reverse'))

    with open(output_path, mode='w') as outfile:
        for path in final_paths:
            outfile.write(path + '\n')

    return



import shutil
import os
import argparse
import glob
from .version import version as VERSION
#VERSION="1.0.0"
import time
from plasmidCC.scripts import utils as utilsCC

from gplas.scripts.node_extraction import extract_nodes, extract_unbinned_solutions
from gplas.scripts.check_prediction_format import check_prediction, PredictionFileFormatError
from gplas.scripts.coverage import coverage
from gplas.scripts.paths import generate_paths
from gplas.scripts.paths_repeats import generate_repeat_paths
from gplas.scripts.coocurrence import calculate_coocurrence
from gplas.scripts.coocurrence_repeats import calculate_coocurrence_repeats
from gplas.scripts.run_plasmidcc import run_plasmidCC, print_speciesopts
from gplas.scripts import utils

start_time = time.time()

# Directories
pkgdir = os.path.dirname(__file__)

# Load ASCII logo
with open(f'{pkgdir}/logo.txt', 'r') as file:
    read_logo = file.read()

#******************************#
#*                            *#
#* Command line parsing       *#
#*                            *#
#******************************#

class PriorityPrinting(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        if option_string == '-h' or option_string == '--help':
            print(read_logo + '\n')
            parser.print_help()
        elif option_string == '-v' or option_string == '--version':
            print(f"gplasCC version {VERSION}")
        elif option_string == '--speciesopts':
            print_speciesopts()
        parser.exit()

#create a function to pass float ranges
parser = argparse.ArgumentParser(description="gplasCC: A tool for binning plasmid-predicted contigs into individual predictions", add_help=False)
parser.register('action', 'printing', PriorityPrinting)

inputgroup = parser.add_argument_group('General')
inputgroup.add_argument('-i', dest='input', type=utils.is_valid_file, required=True, help="Path to the graph file in GFA (.gfa) format, used to extract nodes and links")
inputgroup.add_argument('-o', dest='outdir', type=utils.is_valid_dir, default=".", help="Output directory")  # TODO go through all scripts and add outdir where needed
inputgroup.add_argument('-n', dest='name', type=str, help="Name prefix for output files (default: input file name)")

classifiergroup = inputgroup.add_mutually_exclusive_group(required=True)
classifiergroup.add_argument('-s', dest='species', type=utils.check_species, help="Choose a species database for plasmidCC classification. Use --speciesopts for a list of all supported species")
classifiergroup.add_argument('-p', dest='custom_db_path', type=utilsCC.verify_user_db, help="Path to a custom Centrifuge database (name without file extensions)")
classifiergroup.add_argument('-P', dest='prediction', type=utils.file_exists, help="If not using plasmidCC. Provide a path to an independent binary classification file")
classifiergroup.add_argument('--extract', action='store_true', help="extract FASTA sequences from the assembly graph to use with an external classifier")

paramgroup = parser.add_argument_group('Parameters')
paramgroup.add_argument('-t', dest='threshold_prediction', type=float, default=0.5, help="Prediction threshold for plasmid-derived sequences (default: %(default)s)")
paramgroup.add_argument('-b', dest='bold_coverage_sd', type=int, default=5, help="Coverage variance allowed for bold walks to recover unbinned plasmid-predicted nodes (default: %(default)s)")
paramgroup.add_argument('-r', dest='repeats_coverage_sd', type=int, default=2, help="Coverage variance allowed for assigning repeats to bins (default: %(default)s)")
paramgroup.add_argument('-x', dest='number_iterations', type=int, default=20,help="Number of walk iterations per starting node (default: %(default)s)")
paramgroup.add_argument('-f', dest='filt_gplas', type=float, default=0.1, help="filtering threshold to reject outgoing edges (default: %(default)s)")
paramgroup.add_argument('-e', dest='edge_threshold', type=float, default=0.1, help="Edge threshold (default: %(default)s)")
paramgroup.add_argument('-q', dest='modularity_threshold', type=float, default=0.2, help="Modularity threshold to split components in the plasmidome network (default: %(default)s)")
paramgroup.add_argument('-l', dest='length_filter', type=int, default=1000, help="Filtering threshold for sequence length (default: %(default)s)")

othergroup = parser.add_argument_group('Other')
othergroup.add_argument('-k', '--keep', action='store_true', help="Keep intermediary files")
#othergroup.add_argument('--threads', type=int, default=1, help="Max number of threads to ")  #TODO add multi processing for paths scripts

infogroup = parser.add_argument_group('Info')
infogroup.add_argument('--speciesopts', action='printing', nargs=0, help="Prints a list of all supported species for the -s flag")
infogroup.add_argument('-v', '--version', action='printing', nargs=0, help="Prints gplas version")
infogroup.add_argument('-h', '--help', action='printing', nargs=0, help="Prints this message")
args = parser.parse_args()


#Success Messages
def success_message(outdirname):
    print('\n')
    print(read_logo)
    print(f"""
Congratulations! Prediction succesfully done
Your results are in '{outdirname}'

Thank you for using gplas version {VERSION} we hope it helps your research
Please cite: https://academic.oup.com/bioinformatics/article/36/12/3874/5818483
""")
    end_time = time.time()
    duration = end_time - start_time
    print(f"gplas took {round(duration,1)} seconds to run")
    utils.quit_tool(0)


def success_message_extract():
    print('\n')
    print(read_logo)
    print(f"""
Your nodes have been succesfully extracted to 'gplas_input/{sample}_contigs.fasta'
You can use an external tool to classify these contigs, and then use gplas to bin them into individual plasmids

Thank you for using gplas version {VERSION} we hope it helps your research
Please cite: https://academic.oup.com/bioinformatics/article/36/12/3874/5818483
""")
    utils.quit_tool(0)


#******************************#
#*                            *#
#*        Start gplas         *#
#*                            *#
#******************************#

# Check if the user specified a run name or to use file name data
infile = os.path.abspath(args.input)
infilename = os.path.basename(args.input)
outdirname = os.path.abspath(args.outdir)

if args.name:
    sample = args.name
else:
    sample, _ = os.path.splitext(infilename)

if not args.extract:
    #Print messages
    print('\n')
    print(read_logo)
    print('\n')
    
    #Print chosen parameters
    print("##################################################################")
    print("Your results will be named:...........................", sample)
    print("Input graph:..........................................", infilename)
    print("Threshold for predicting plasmid-derived contigs:.....", args.threshold_prediction)
    print("Number of plasmid walks created per node:.............", args.number_iterations)
    print("Threshold of gplas scores:............................", args.filt_gplas)
    print("Minimum frequency to consider an edge:................", args.edge_threshold)
    print("Modularity threshold used to partition the network:...", args.modularity_threshold)
    print("Coverage SD for bold mode:............................", args.bold_coverage_sd)
    print("Minimum sequence length:..............................", args.length_filter)
    print("##################################################################" + '\n')

    print("Extracting contigs from the assembly graph...", end='\r')

##_1.0 Run analysis
#_1.1 Extract nodes and links from the assembly graph
os.makedirs(f'{outdirname}/gplas_input', exist_ok=True)

extract_nodes(sample, infile, args.length_filter,outdirname)
utils.check_output(f"{outdirname}/gplas_input/{sample}_raw_nodes.fasta")

#_1.2 If in extract mode, exit workflow after succesful extraction. Else continue workflow
if args.extract:
    success_message_extract() #Exits the workflow
print("Extracting contigs from the assembly graph............ completed!")

##_2.0 Obtain correct prediction file
#_2.1 Run plasmidCC if no independent prediction file was given
if args.species or args.custom_db_path:
    print("Running plasmidCC to generate prediction file:" + '\n')
    os.makedirs(f'{outdirname}/plasmidCC', exist_ok=True)
    inputFASTA = f"{outdirname}/gplas_input/{sample}_contigs.fasta"

    run_plasmidCC(inputFASTA, sample, args.length_filter, args.species, args.custom_db_path,outdirname)
    utils.cleanup_centrifuge(sample,outdirname)

    print('\n', end='')
    path_prediction = f"{outdirname}/plasmidCC/{sample}/{sample}_gplas.tab"
else:
    path_prediction = args.prediction

utils.check_output(path_prediction)

#_2.2 Check if the prediction file is correctly formatted.
print("Checking prediction file...", end='\r')

try:
    check_prediction(sample, path_prediction,outdirname)
except PredictionFileFormatError as err:
    print('\n\n' + "Error in prediction file format:")
    print(err)
    utils.quit_tool(err)
print("Checking prediction file.............................. completed!")

##_3.0 Run gplas in normal mode
#_3.1 Extract nodes/links from the assembly graph
print("Processing input data...", end='\r')
os.makedirs(f'{outdirname}/coverage', exist_ok=True)

coverage(sample, path_prediction, args.threshold_prediction, outdirname)
print("Processing input data................................. completed!")

# Check for suitable plasmid nodes
init_nodes_path = f"{outdirname}/coverage/{sample}_initialize_nodes.tab"
with open(init_nodes_path, mode='r') as file:
    line_content = file.readline()
if not line_content:
    print("There are no suitable plasmids to initiate a random walk. gplas can't do anything")
    if not args.keep:
        utils.cleanup_intermediary_files(sample,outdirname)
    utils.quit_tool(-1)

#_3.2 Generate random walks
print("Generating random walks in normal mode...", end='\r')
os.makedirs(f"{outdirname}/walks/normal_mode", exist_ok=True)

generate_paths(sample, args.number_iterations, args.filt_gplas, outdirname, mode='normal')
print("Generating random walks in normal mode................ completed!")

#_3.3 Calculate coocurrence between walks
print("Calculating coocurrence of random walks...", end='\r')
os.makedirs(f"{outdirname}/results/normal_mode", exist_ok=True)

if not calculate_coocurrence(sample, args.number_iterations, args.threshold_prediction, args.modularity_threshold, outdirname, mode='normal'):
    print("Calculating coocurrence of random walks............... completed!")
    print("gplas couldn't find any walks connecting plasmid-predicted nodes")
    print("Plasmid nodes will be classified as Unbinned. If this is unexpected, please assemble your genome with different parameters or with a different tool and re-run gplas")
else:
    print("Calculating coocurrence of random walks............... completed!")
utils.check_output(f"{outdirname}/results/normal_mode/{sample}_results_no_repeats.tab")

##_4.0 Resolve unbinned contigs
#_4.1 Check for unbinned contigs
unbinned_path = f"{outdirname}/results/normal_mode/{sample}_bin_Unbinned.fasta"
if os.path.exists(unbinned_path):
    #_4.1.1 Run gplas in bold mode if contigs were left unbinned
    print("Some contigs were left unbinned")  # improve tell user how many contigs are unbinned?
    #_4.1.1.1 Generate random walks
    print("Generating random walks in bold mode...", end='\r')
    os.makedirs(f"{outdirname}/walks/bold_mode", exist_ok=True)

    generate_paths(sample, args.number_iterations, args.filt_gplas, outdirname,args.bold_coverage_sd, mode='bold')
    print("Generating random walks in bold mode.................. completed!")

    #_4.1.1.2 Extract unbinned solutions
    os.makedirs(f"{outdirname}/walks/unbinned_nodes", exist_ok=True)
    extract_unbinned_solutions(sample, outdirname)

    #_4.1.1.3 Recalculate coocurrence of walks using the combined solutions
    print("Recalculating coocurrence of random walks...", end='\r')

    if not calculate_coocurrence(sample, args.number_iterations, args.threshold_prediction, args.modularity_threshold, outdirname, mode='unbinned'):
        print("Recalculating coocurrence of random walks............. completed!")
        print("gplas bold mode couldn't find any walks connecting plasmid-predicted nodes")
        print("Plasmid nodes will be classified as Unbinned. If this is unexpected, please assemble your genome with different parameters or with a different tool and re-run gplas")
    else:
        print("Recalculating coocurrence of random walks............. completed!")

#_4.1.2 Copy files from normal mode if there were no unbinned contigs
else:
    for file in glob.glob(f"{outdirname}/results/normal_mode/{sample}*"):
        shutil.copy(file, f"{outdirname}/results/")

utils.check_output(f"{outdirname}/results/{sample}_results_no_repeats.tab")

##_5.0 Add repeated elements
#_5.1 Check for repeats
repeated_elements_path = f"{outdirname}/coverage/{sample}_repeat_nodes.tab"
with open(repeated_elements_path, mode='r') as file:
    line_content = file.readline()
if line_content:
    #_5.1.1 Run gplas on repeated elements
    print("Adding repeated elements to the predictions...", end='\r')
    os.makedirs(f"{outdirname}/walks/repeats", exist_ok=True)

    #_5.1.1.1 Generate random walks
    generate_repeat_paths(sample, args.number_iterations, args.filt_gplas, outdirname)

    #_5.1.1.2 Calculate coocurrence between walks
    if not calculate_coocurrence_repeats(sample, outdirname, args.repeats_coverage_sd):
        print("Adding repeated elements to the predictions........... completed!")
        print("gplas couldn't find any walks connecting repeats to plasmid-nodes")
        shutil.move(f"{outdirname}/results/{sample}_results_no_repeats.tab", f"results/{sample}_results.tab")
        shutil.move(f"{outdirname}/results/{sample}_bins_no_repeats.tab", f"results/{sample}_bins.tab")

#_5.1.2 If there are no repeated elements, just rename the results files
else:
    shutil.move(f"{outdirname}/results/{sample}_results_no_repeats.tab", f"{outdirname}/results/{sample}_results.tab")
    shutil.move(f"{outdirname}/results/{sample}_bins_no_repeats.tab", f"{outdirname}/results/{sample}_bins.tab")

utils.check_output(f"{outdirname}/results/{sample}_results.tab")

##_6.0 If the -k flag was not selected, delete intermediary files
if not args.keep:
    utils.cleanup_intermediary_files(sample,outdirname)

##_7.0 Show success message and exit workflow
success_message(outdirname)
